@@ -92,9 +92,11 @@ enum {
 	PAT_MATCH_NUM
 };
 
-#define PAT_REF_MAP 0x1 /* Set if the reference is used by at least one map. */
-#define PAT_REF_ACL 0x2 /* Set if the reference is used by at least one acl. */
-#define PAT_REF_SMP 0x4 /* Flag used if the reference contains a sample. */
+#define PAT_REF_MAP  0x01 /* Set if the reference is used by at least one map. */
+#define PAT_REF_ACL  0x02 /* Set if the reference is used by at least one acl. */
+#define PAT_REF_SMP  0x04 /* Flag used if the reference contains a sample. */
+#define PAT_REF_FILE 0x08 /* Set if the reference was loaded from a file */
+#define PAT_REF_ID   0x10 /* Set if the reference is only an ID (not loaded from a file) */
 
 /* This struct contain a list of reference strings for dunamically
  * updatable patterns.
@@ -205,7 +207,6 @@ struct pattern_expr {
 	struct eb_root pattern_tree;  /* may be used for lookup in large datasets */
 	struct eb_root pattern_tree_2;  /* may be used for different types */
 	int mflags;                     /* flags relative to the parsing or matching method. */
-	uint32_t refcount;            /* refcount used to know if the expr can be deleted or not */
 	__decl_thread(HA_RWLOCK_T lock);               /* lock used to protect patterns */
 };
 
@@ -215,6 +216,7 @@ struct pattern_expr {
  */
 struct pattern_expr_list {
 	struct list list; /* Used for chaining pattern_expr in pattern_head. */
+	int do_free;
 	struct pattern_expr *expr; /* The used expr. */
 };
 
