@@ -47,9 +47,9 @@
 #define TASK_WOKEN_TIMER  0x00000200  /* woken up because of expired timer */
 #define TASK_WOKEN_IO     0x00000400  /* woken up because of completed I/O */
 #define TASK_WOKEN_SIGNAL 0x00000800  /* woken up by a system signal */
-#define TASK_WOKEN_MSG    0x00001000  /* woken up by another task's message (see also UEVT/USR1) */
+#define TASK_WOKEN_MSG    0x00001000  /* woken up by another task's message */
 #define TASK_WOKEN_RES    0x00002000  /* woken up because of available resource */
-#define TASK_WOKEN_OTHER  0x00004000  /* woken up for an unspecified reason (see also UEVT/USR1) */
+#define TASK_WOKEN_OTHER  0x00004000  /* woken up for an unspecified reason */
 
 /* use this to check a task state or to clean it up before queueing */
 #define TASK_WOKEN_ANY    (TASK_WOKEN_OTHER|TASK_WOKEN_INIT|TASK_WOKEN_TIMER| \
@@ -58,9 +58,7 @@
 
 #define TASK_F_TASKLET    0x00008000  /* nature of this task: 0=task 1=tasklet */
 #define TASK_F_USR1       0x00010000  /* preserved user flag 1, application-specific, def:0 */
-#define TASK_F_UEVT1      0x00020000  /* one-shot user event type 1, application specific, def:0 */
-#define TASK_F_UEVT2      0x00040000  /* one-shot user event type 2, application specific, def:0 */
-/* unused: 0x80000..0x80000000 */
+/* unused: 0x20000..0x80000000 */
 
 /* These flags are persistent across scheduler calls */
 #define TASK_PERSISTENT   (TASK_SELF_WAKING | TASK_KILLED | \
@@ -166,6 +164,24 @@ struct tasklet {
 	 */
 };
 
+/* Note: subscribing to these events is only valid after the caller has really
+ * attempted to perform the operation, and failed to proceed or complete.
+ */
+enum sub_event_type {
+	SUB_RETRY_RECV       = 0x00000001,  /* Schedule the tasklet when we can attempt to recv again */
+	SUB_RETRY_SEND       = 0x00000002,  /* Schedule the tasklet when we can attempt to send again */
+};
+
+/* Describes a set of subscriptions. Multiple events may be registered at the
+ * same time. The callee should assume everything not pending for completion is
+ * implicitly possible. It's illegal to change the tasklet if events are still
+ * registered.
+ */
+struct wait_event {
+	struct tasklet *tasklet;
+	int events;             /* set of enum sub_event_type above */
+};
+
 /*
  * The task callback (->process) is responsible for updating ->expire. It must
  * return a pointer to the task itself, except if the task has been deleted, in
