@@ -230,7 +230,7 @@ static void _do_poll(struct poller *p, int exp, int wake)
 		int timeout = (global.tune.options & GTUNE_BUSY_POLLING) ? 0 : wait_time;
 
 		status = epoll_wait(epoll_fd[tid], epoll_events, global.tune.maxpollevents, timeout);
-		clock_update_local_date(wait_time, (global.tune.options & GTUNE_BUSY_POLLING) ? 1 : status);
+		clock_update_local_date(timeout, status);
 
 		if (status) {
 			activity[tid].poll_io++;
@@ -275,6 +275,8 @@ static int init_epoll_per_thread()
 	epoll_events = calloc(1, sizeof(struct epoll_event) * global.tune.maxpollevents);
 	if (epoll_events == NULL)
 		goto fail_alloc;
+	vma_set_name_id(epoll_events, sizeof(struct epoll_event) * global.tune.maxpollevents,
+	                "ev_epoll", "epoll_events", tid + 1);
 
 	if (MAX_THREADS > 1 && tid) {
 		epoll_fd[tid] = epoll_create(global.maxsock + 1);
