@@ -655,6 +655,7 @@ int
 flt_http_payload(struct stream *s, struct http_msg *msg, unsigned int len)
 {
 	struct filter *filter;
+	struct htx *htx;
 	unsigned long long *strm_off = &FLT_STRM_OFF(s, msg->chn);
 	unsigned int out = co_data(msg->chn);
 	int ret, data;
@@ -670,17 +671,19 @@ flt_http_payload(struct stream *s, struct http_msg *msg, unsigned int len)
 		/* Call http_payload for filters only. Forward all data for
 		 * others and update the filter offset
 		 */
-		if (!IS_DATA_FILTER(filter, msg->chn) || !FLT_OPS(filter)->http_payload) {
+		if (!IS_DATA_FILTER(filter, msg->chn)) {
 			*flt_off += data - offset;
 			continue;
 		}
 
-		DBG_TRACE_DEVEL(FLT_ID(filter), STRM_EV_HTTP_ANA|STRM_EV_FLT_ANA, s);
-		ret = FLT_OPS(filter)->http_payload(s, filter, msg, out + offset, data - offset);
-		if (ret < 0)
-			goto end;
-		data = ret + *flt_off - *strm_off;
-		*flt_off += ret;
+		if (FLT_OPS(filter)->http_payload) {
+			DBG_TRACE_DEVEL(FLT_ID(filter), STRM_EV_HTTP_ANA|STRM_EV_FLT_ANA, s);
+			ret = FLT_OPS(filter)->http_payload(s, filter, msg, out + offset, data - offset);
+			if (ret < 0)
+				goto end;
+			data = ret + *flt_off - *strm_off;
+			*flt_off += ret;
+		}
 	}
 
 	/* If nothing was forwarded yet, we take care to hold the headers if
@@ -700,6 +703,9 @@ flt_http_payload(struct stream *s, struct http_msg *msg, unsigned int len)
 	ret = data;
 	*strm_off += ret;
  end:
+	htx = htxbuf(&msg->chn->buf);
+	if (msg->flags & HTTP_MSGF_XFER_LEN)
+		htx->extra = 0;
 	DBG_TRACE_LEAVE(STRM_EV_STRM_ANA|STRM_EV_HTTP_ANA|STRM_EV_FLT_ANA, s);
 	return ret;
 }
@@ -950,17 +956,20 @@ flt_tcp_payload(struct stream *s, struct channel *chn, unsigned int len)
 		/* Call tcp_payload for filters only. Forward all data for
 		 * others and update the filter offset
 		 */
-		if (!IS_DATA_FILTER(filter, chn) || !FLT_OPS(filter)->tcp_payload) {
+		if (!IS_DATA_FILTER(filter, chn)) {
 			*flt_off += data - offset;
 			continue;
 		}
 
-		DBG_TRACE_DEVEL(FLT_ID(filter), STRM_EV_TCP_ANA|STRM_EV_FLT_ANA, s);
-		ret = FLT_OPS(filter)->tcp_payload(s, filter, chn, out + offset, data - offset);
-		if (ret < 0)
-			goto end;
-		data = ret + *flt_off - *strm_off;
-		*flt_off += ret;
+		if (FLT_OPS(filter)->tcp_payload) {
+
+			DBG_TRACE_DEVEL(FLT_ID(filter), STRM_EV_TCP_ANA|STRM_EV_FLT_ANA, s);
+			ret = FLT_OPS(filter)->tcp_payload(s, filter, chn, out + offset, data - offset);
+			if (ret < 0)
+				goto end;
+			data = ret + *flt_off - *strm_off;
+			*flt_off += ret;
+		}
 	}
 
 	/* Only forward data if the last filter decides to forward something */
