@@ -206,8 +206,9 @@ static int quic_conn_enc_level_init(struct quic_conn *qc,
 	if (!qel)
 		goto leave;
 
-	LIST_INIT(&qel->retrans);
-	qel->retrans_frms = NULL;
+	LIST_INIT(&qel->el_send);
+	qel->send_frms = NULL;
+
 	qel->tx.crypto.bufs = NULL;
 	qel->tx.crypto.nb_buf = 0;
 	qel->cstream = NULL;
@@ -243,23 +244,7 @@ static int quic_conn_enc_level_init(struct quic_conn *qc,
 			goto err;
 	}
 
-	/* Ensure early-data encryption is not inserted at the end of this ->qel_list
-	 * list. This would perturbate the sender during handshakes. This latter adds
-	 * PADDING frames to datagrams from the last encryption level in this list,
-	 * for datagram with at least an ack-eliciting Initial packet inside.
-	 * But a QUIC server has nothing to send from this early-data encryption
-	 * level, contrary to the client.
-	 * Here early-data is added after the Initial encryption level which is
-	 * always already present.
-	 */
-	if (level == ssl_encryption_early_data) {
-		if (qc->iel)
-			LIST_APPEND(&qc->iel->list, &qel->list);
-		else
-			LIST_INSERT(&qc->qel_list, &qel->list);
-	}
-	else
-		LIST_APPEND(&qc->qel_list, &qel->list);
+	LIST_APPEND(&qc->qel_list, &qel->list);
 	*el = qel;
 	ret = 1;
  leave:
